#include <Arduino.h>
#include <Wire.h>
#include <map>
#include <vector>
#include <string>

#include "BMP280Sensor.h"
#include "CCS811Sensor.h"
#include "HDC1080Sensor.h"



std::map<ReadingType, std::map<std::string, float>> sensorStore;
typedef std::map<uint8_t, i2cSensor*> SensorMap;

// Declare sensors
BMP280Sensor bmp280;
CCS811Sensor ccs811;
HDC1080Sensor hdc1080;

// global values to make CCS811Sensor and HDC1080Sensor share data
double sharedTemperature = NAN;
double sharedHumidity = NAN;


// Map to store sensor instances by their I2C addresses
SensorMap i2cSensors;

// Map to store initialized sensors by their names
std::map<std::string, i2cSensor*> initializedi2cSensors;

// Function to initialize sensor store
void initSensorStore() {
  for (ReadingType type : {ReadingType::Temperature, ReadingType::Humidity, ReadingType::Pressure, ReadingType::Altitude, ReadingType::eCO2, ReadingType::eTVOC, ReadingType::ErrorStatus}) {
    sensorStore[type] = std::map<std::string, float>();
  }
}

// Function to scan for I2C devices
std::vector<uint8_t> scanI2C() {
  std::vector<uint8_t> devices;
  for (uint8_t address = 1; address < 127; ++address) {
    Wire.beginTransmission(address);
    if (Wire.endTransmission() == 0) {
      devices.push_back(address);
    }
  }
  return devices;
}

void initializeSensors(const std::vector<uint8_t>& i2cDevices, bool specialInit = false) {
  for (uint8_t address : i2cDevices) {
    // Check if the address matches any known sensors
    if (i2cSensors.find(address) != i2cSensors.end()) {
      i2cSensor* sensor = i2cSensors[address];
      std::string output = sensor->getName();

      if (!specialInit) {
        // First pass: skip sensors that require special initialization
        if (sensor->getName() == "CCS811") {
          continue;
        }
        output = "Regualr " + output;
        // Default initialization for all other sensors
        sensor->init();
      } else if (sensor->getName() == "CCS811") {
        output = "Special " + output;
        // Initialize CCS811 with references to shared temperature and humidity
        CCS811Sensor* ccs811Sensor = static_cast<CCS811Sensor*>(sensor);
        ccs811Sensor->init(&sharedTemperature, &sharedHumidity);
      }
       Serial.println(output.c_str());
      // Store the initialized sensor
      initializedi2cSensors[sensor->getName()] = sensor;
    }
  }
}

// Main setup function
void setup() {
  Serial.begin(115200);
  Wire.begin();
  delay(10000);
  // Initialize the sensor store
  initSensorStore();

  // Populate the i2cSensors map
  i2cSensors[bmp280.getAddress()] = &bmp280;
  i2cSensors[hdc1080.getAddress()] = &hdc1080;
  i2cSensors[ccs811.getAddress()] = &ccs811;

  // Scan for I2C devices
  std::vector<uint8_t> i2cDevices = scanI2C();

  // First pass: Initialize all sensors with default init()
  Serial.println("First initialization pass...");
  initializeSensors(i2cDevices, false);

  // Optionally, wait or perform some task to allow initial data collection
  delay(5000); // Adjust delay as necessary

  // Second pass: Re-initialize sensors with special requirements
  Serial.println("Second initialization pass...");
  initializeSensors(i2cDevices, true);
}

// Main loop function
void loop() {
  // Example: Read and store values asynchronously
  for (auto& sensorPair : initializedi2cSensors) {
    i2cSensor* sensor = sensorPair.second;
    for (ReadingType feature : sensor->getFeatures()) {
      float value = sensor->readValue(feature);
      sensorStore[feature][sensor->getName()] = value;

      // Update shared variables if HDC1080 is read
      if (sensor->getName() == "HDC1080 Sensor") {
        if (feature == ReadingType::Temperature) {
          sharedTemperature = static_cast<double>(value);
        } else if (feature == ReadingType::Humidity) {
          sharedHumidity = static_cast<double>(value);
        }
      }

      std::string output = sensor->getName() + " " + readingTypeToString(feature) + ": " + std::to_string(value);
      Serial.println(output.c_str());
    }
  }

  delay(20000); // Delay between readings
}
